/***************************************************************************
                          ezconfig  -  description
                             -------------------
    begin                : Dec. 26, 2004
    copyright            : (C) 2005 by Allen
    email                : bon_ami_@hotmail.com
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   Explicit Distribution Limitation                                      *
 *   This rule overrides others below.                                     *
 *   This program may not be modified or used by, or, if possible,         *
 *   redistributed to people described as below,                           *
 *   1.Japanese who hold hostility against Chinese.                        *
 *   2.or, those who discriminate against people based solely on race,     *
 *     gender or sexual orientation.                                       *
 *                                                                         *
 ***************************************************************************/
/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/
/*
 *	Configuration server and client code file
 */

#include <fstream>
#include <sstream>

#include "ezconfig.h"

namespace ezproject {

/*	search for a vector in this vector's queue & sub-queues,
	comparing ID or name.
Input Parameters:
	dp:	the IDs or names of the vectors on the way to the one asked for.
		the upper its position is, or, the closer it's to the root, the latter
		it is in the list.
		cleared if found. the ones remained in the list help keep track of
		where a vector cannot be found.
	vc:	set if found. otherwise, not set.
Return Values:
	RTTP_CRER:	input vc cannot be returned
	RTTP_MNER:	no such vector. the last vector in the vector list is the
				one not found
	RTTP_RGHT:	sub-vector found.
*/
erttp ezconfig::cfgvec::getvec(vector<vctrdpttp> &dp, cfgvec *&vc) const
{
	if (0 == dp.size())
	{
		vc = const_cast<cfgvec *>(this);
		return RTTP_RGHT;	//this is it, i.e. all criteria satisfied
	}

	vctrdpttp sch = dp.back();	//the searching criteria

	if (sch.first != 0)	//search by ID
	{
		vctrmnttp id;

		for (vector<cfgvec>::const_iterator it = m_q.begin(); it != m_q.end();
			it++)
		{
			//search leaves / brances / the queue
			it->getid(id);
			if (sch.first == id)
			{
				dp.pop_back();	//criteria satisfied once
				return it->getvec(dp, vc);	//more criteria?
			}
		}
	}
	else	//search by name
	{
		string st;

		for (vector<cfgvec>::const_iterator it = m_q.begin(); it != m_q.end();
			it++)
		{
			it->getnm(st);
			if (0 == sch.second.compare(st))
			{
				dp.pop_back();
				return it->getvec(dp, vc);
			}
		}
	}
	return RTTP_MNER;	//not found
}

/*	sub-vector registration
Input Parameters:
	id: value omitted, set if successful.
	nm: set to vector if successful.
Return Values:
	RTTP_CRER:	queue full
	RTTP_NOER:	there's at least one hole in the queue. defragmentation needed
	RTTP_RGHT:	sub-vector registered.
*/
erttp ezconfig::cfgvec::regq(vctrmnttp &id, string nm)
{
	erttp ret = RTTP_NA;
	vctrmnttp mx = 0;
	vector<cfgvec>::iterator it = m_q.begin();

	if (m_q.size() > 0)
	{
		m_q.back().getid(id);
		mxvl(mx);
	}
	else
		id = 0; //the 1st. vector
	if (mx == ++id)
	{
		vctrmnttp pid = 0;

		ret = RTTP_CRER;	//full? id is never equal to 0
		for (; it != m_q.end(); it++)
		{
			it->getid(id);
			if (++pid != id)
			{
				//gap
				ret = RTTP_NOER;
				id = pid;
				break;
			}
		}
	}
	if (ret > RTTP_MNER)
	{
		cfgvec vecsmpl; //vector sample

		vecsmpl.setid(id);
		vecsmpl.setnm(nm);
		m_q.insert(it, vecsmpl);
		vctrdpttp idnm = make_pair(0, "");
		wlk(idnm);
		if (RTTP_NA == ret)
			ret = RTTP_RGHT;
	}
	return ret;
}

erttp ezconfig::cfgvec::unregq(vector<vctrdpttp> &dp)
{
	erttp ret = RTTP_MNER;

	if (dp.size() < 1)
		ret = RTTP_CRER;
	else
	{
		vctrdpttp sch = dp.at(0);
		vector<cfgvec>::iterator it = m_q.begin();

		if (sch.first != 0) //search by ID
		{
			vctrmnttp id;

			for (; it != m_q.end(); it++)
			{
				it->getid(id);
				if (sch.first == id)
				{
					ret = RTTP_RGHT;
					break;
				}
			}
		}
		else
		{
			string st;

			for (; it != m_q.end(); it++)
			{
				it->getnm(st);
				if (0 == sch.second.compare(st))
				{
					ret = RTTP_RGHT;
					break;
				}
			}
		}
		if (ret >= RTTP_NOER)
		{
			m_q.erase(it);
			vctrdpttp idnm = make_pair(0, "");
			wlk(idnm);
		}
	}
	return ret;
}

/*	sub-elements renumber
Input/Output Parameters:
	ot:	renumbered ID, name pair of sub-elements
	old:	IDs before renumber, for comparision to the new ones
	dp:	element tree to locate which elements to renumber
Return Values:
	RTTP_NOER:	full
	RTTP_RGHT:	renumbered
*/
erttp ezconfig::cfgvec::sqz(vector<vctrdpttp> &ot, vector<vctrmnttp> &old)
{
	erttp ret = RTTP_RGHT;
	vector<cfgvec>::iterator it = m_q.begin();
	vctrmnttp sub = 1, id;
	vctrdpttp idnm;
	string nm;

	for (; it != m_q.end(); it++)
	{
		it->getid(id);
		it->getnm(nm);
		old.push_back(id);
		it->setid(sub);
		idnm = make_pair(sub, nm);
		sub++;
		ot.push_back(idnm);
	}
	idnm.first = 0;
	idnm.second.clear();
	wlk(idnm);
	if (sub <= 0)
		ret = RTTP_NOER;
	return ret;
}

/*	walk over sub-elements
Input/Output Parameters:
	nf:	vector ID & name returned, if 0 & empty input, 1st. vector's returned
Return Values:
	RTTP_MNER:	no vectors
	RTTP_NOER:	1st. vector got
	RTTP_RGHT:	vector got or walk reset
*/
erttp ezconfig::cfgvec::wlk(vctrdpttp &nf) const
{
	static vector<cfgvec>::const_iterator vecsttc = m_q.end();
	erttp ret = RTTP_RGHT;

	if (m_q.size() < 1)
		ret = RTTP_MNER;
	else
	{
		if ((0 == nf.first) && nf.second.empty())
			vecsttc = m_q.end();
		else
		{
			if (m_q.end() == vecsttc)
			{
				vecsttc = m_q.begin();
				ret = RTTP_NOER;
			}
			else
				vecsttc++;
			if (m_q.end() == vecsttc)
				ret = RTTP_MNER;
			else
			{
				vecsttc->getid(nf.first);
				vecsttc->getnm(nf.second);
			}
		}
	}
	return ret;
}


/*	standard configuation client.
Input/Output Parameters:
	fn:	file name (absolute/relative/no path name);
	duphndl:	duplicate items handling method;
	errmsg:	error message
Return Values:
	RTTP_CRER:	file opening failure, and no help interface provided,
				or, no item handling interface provided
	RTTP_MNER:	file opening failure, and exit as help interface requested,
	RTTP_NOER:	exit as item interface requested
	RTTP_RGHT:	section(s)/item(s) recognized, and exit reaching the end
	RTTP_NA:	no items recognized
*/
erttp ezcnfgclnt::run(string &fn, string &errmsg)
{
	std::ifstream fs;
	string sctn;
	string buf;
	size_t bufpos;
	std::stringstream ss;
	unsigned int lineno = 0;
	int res;
	erttp ret = RTTP_NA;
	bool rt;

	if (NULL == itmf)
	{
		ret = RTTP_CRER;
		errmsg.append("no config handling interface found. "
			"file opening terminated\n");
	}
	while (RTTP_NA == ret)
	{
		if (fn.empty())
			res = -1;
		else
		{
			fs.open(fn.c_str());
			if (fs.good())
				res = 0;
			else
				res = -1;
		}
		if (res != 0)
		{
			errmsg.append("\"");
			errmsg.append(fn);
			errmsg.append("\" opening failed\n");
			if (NULL == hlp)
			{
				ret = RTTP_CRER;
				errmsg.append("file opening terminated\n");
			}
			else
				if (!hlp(fn))
				{
					ret = RTTP_MNER;
					errmsg.append("file opening aborted\n");
				}
				else
					fs.clear();
		}
		else
			ret = RTTP_NOER;
	}
	if (ret < RTTP_NOER)
		return ret;
	ret = RTTP_NA;
	while (getline(fs, buf, '\n'))
	{
		lineno++;
		if (buf.empty() || ('#' == buf.at(0)) || ('\0' == buf.at(0))
			|| (('`' == buf.at(0)) && (buf.length() < 2)))
			//empty or comment line
			continue;
		if (('`' == buf.at(0)) && ('`' != buf.at(1)))
			//section label
			sctn = buf.substr(1);
		else
		{
			//section content
			if ('`' == buf.at(0))
				buf.erase(0, 1);
			if (buf.length() > 2)
			{
				if (' ' == buf.at(1))
					//line is "[a single character] [value]"
					rt = itmf(sctn, buf.substr(0, 1), buf.substr(2));
				else
					if (string::npos != (bufpos = buf.find('=', 0)))
						//line is "[item name]=[value]"
						rt = itmf(sctn,
							buf.substr(0, bufpos),
							buf.substr(bufpos + 1));
					else
						//line is "[item to be set]"
						rt = itmf(sctn, buf, "");
			}
			else
				//single character
				rt = itmf(sctn, buf, "");
			if (RTTP_NA == ret)
				ret = RTTP_RGHT;
			if (!rt)
			{
				ret = RTTP_NOER;
				errmsg.append("file parsing aborted at line ");
				ss << lineno << '\n';
				errmsg.append(ss.str());
				break;
			}
		}
	}

	return ret;
}

/*	standard configuation client.
Input/Output Parameters:
	argc:	command argument count;
	argv:	command argument values;
	errmsg:	error message
Return Values:
	RTTP_CRER:	no item handling interface provided
	RTTP_MNER:
	RTTP_NOER:	exit as item interface requested
	RTTP_RGHT:	section(s)/item(s) recognized, and exit reaching the end
	RTTP_NA:	no items recognized
*/
erttp ezcnfgclnt::run(int argc, char *argv[], string &errmsg)
{
	ezoi oi(&std::cin, &std::cout, &std::cerr);
	int i, j, sctn;
	string vl;
	erttp ret = RTTP_NA;

	if (NULL == itml)
	{
		ret = RTTP_CRER;
		errmsg.append("no config handling interface found. "
			"command line parsing terminated\n");
	}
	else
		for (i = 1; i < argc; i++)
		{
			if ('-' == *argv[i])
			{
				switch (argv[i][sctn = 1])
				{
					case '-':
						if ('\0' != argv[i][sctn = 2])
							break;
						/* fall through */
					case '\0':
						errmsg.append("Error parsing '");
						errmsg.append(argv[i]);
						errmsg.append("'\n");
						continue;
				}
				/* -x or --x found. search for its value */
				vl.clear();
				for (j = i++; (i < argc) && ('-' != *argv[i]); i++)
					vl.append(argv[i]);	//value found
				if (!vl.empty())
				{
					/* value found */
					ret = RTTP_RGHT;
					if (!itml(&argv[j][sctn], vl))
					{
						ret = RTTP_NOER;
						break;
					}
				}
			}
		}

	return ret;
}

}	/*	namespace ezproject	*/

