/***************************************************************************
                          ezconfig  -  description
                             -------------------
    begin                : Dec. 26, 2004
    copyright            : (C) 2005 by Allen
    email                : bon_ami_@hotmail.com
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   Explicit Distribution Limitation                                      *
 *   This rule overrides others below.                                     *
 *   This program may not be modified or used by, or, if possible,         *
 *   redistributed to people described as below,                           *
 *   1.Japanese who hold hostility against Chinese.                        *
 *   2.or, those who discriminate against people based solely on race,     *
 *     gender or sexual orientation.                                       *
 *                                                                         *
 ***************************************************************************/
/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/
/*
 *	header of EZ Config
 *
 *	configuration server and client for all kinds of configurations
 */

#ifndef EZCONFIG_H
#define EZCONFIG_H

#include <vector>
#include <utility>
#include <iostream>
#include "ezcommon.h"

namespace ezproject {

using std::vector;
using std::pair;
using std::make_pair;
using std::string;

/*	calculate the maximum positive value a type can hold.
	only whole-bytes possession types can be handled.
	warning saying the evaluation inside is always true during compilation
	may safely be ignored */
template <typename Type>
void mxvl(Type &vl)
{
	memset(&vl, 0xff, sizeof(vl));
	if (vl < 0)
	{
		vl = 0x1;
		while (vl > 0)
			vl <<= 1;
		vl = ~vl;
	}
}

class ezconfig	//configuration server
{
public:
	typedef unsigned char vctrlvltp;	//vector level type
	typedef unsigned char vctrmnttp;	//vector amount type
	typedef pair<vctrmnttp, string> vctrdpttp;	//vector (along a tree)
		//depth amount type

private:
	class cfgvec	//configuration vector
	{
	private:
		vctrmnttp m_vecid;	//vector ID
		string m_vecnm;	//vector name
		void *m_vecvl;	//vector value
		template <class EleType = string>
		class cfgele	//vector element (value), levels are from 0
		{
		private:
			typedef pair<bool, EleType> elevltp;	//element whether set & its
			// value
			vector<elevltp> m_elevl;	//element value

			erttp getbsc(vctrlvltp lvl, EleType &vl) const;

		public:
			erttp reg(const vctrlvltp HowManyLevels);
			erttp set(const vctrlvltp level, const EleType value);
			erttp get(vctrlvltp &lvl, EleType &vl) const;
		};

	public:
		vector<cfgvec> m_q; //sub-vectors / leaves / branches

		cfgvec(void) { m_vecvl = NULL; }
		/* there is no way to destruct it voluntarily */

		erttp getvec(vector<vctrdpttp> &VectorTreeDepth, cfgvec *&VectorFound)
			const;

		template <typename EleType>
		erttp regele(vctrlvltp HowManyLevels);
		template <typename EleType>
		erttp unregele(void);
		void setid(const vctrmnttp VectorID) { m_vecid = VectorID; }
		void getid(vctrmnttp &VectorID) const { VectorID = m_vecid; }
		template <typename EleType>
		erttp setele(vctrlvltp level, EleType value);
		template <typename EleType>
		erttp getele(vctrlvltp &level, EleType &value, string &name) const;
		void setnm(const string &VectorName) { m_vecnm.assign(VectorName); }
		void getnm(string &VectorName) const { VectorName.assign(m_vecnm); }

		erttp regq(vctrmnttp &SubVectorID, string VectorName);
		erttp unregq(vector<vctrdpttp> &VectorID);
		erttp sqz(vector<vctrdpttp> &OutputVectors,
			vector<vctrmnttp> &OldVectors);
		erttp wlk(vctrdpttp &VectorInfo) const;
	};
	cfgvec m_rt;	//root

public:
	template <typename EleType>
	erttp regele(vector<vctrdpttp> &VectorTreeDepth, vctrlvltp HowManyLevels)
		{ cfgvec *vec;
			if (m_rt.getvec(VectorTreeDepth, vec) < RTTP_NOER) return RTTP_MNER;
			return vec->regele<EleType>(HowManyLevels); }
	template <typename EleType>
	erttp unregele(vector<vctrdpttp> &VectorTreeDepth, vctrdpttp &VectorInfo);
	template <typename EleType>
	erttp setele(vector<vctrdpttp> &VectorTreeDepth, vctrlvltp level,
		EleType value)
		{ cfgvec *vec;
			if (m_rt.getvec(VectorTreeDepth, vec) < RTTP_NOER) return RTTP_MNER;
			return vec->setele(level, value); }
	template <typename EleType>
	erttp getele(vector<vctrdpttp> &VectorTreeDepth, vctrlvltp &level,
		EleType &value, string &name) const
		{ cfgvec *vec;
			if (m_rt.getvec(VectorTreeDepth, vec) < RTTP_NOER) return RTTP_MNER;
			return vec->getele(level, value, name); }
	erttp nmele(vector<vctrdpttp> &VectorTreeDepth, const string NewName,
		string &OldName)
		{ cfgvec *vec;
			if (m_rt.getvec(VectorTreeDepth, vec) < RTTP_NOER) return RTTP_MNER;
			vec->getnm(OldName); vec->setnm(NewName); return RTTP_RGHT; }
	erttp nmele(vector<vctrdpttp> &VectorTreeDepth, string &OldName)
		{ cfgvec *vec;
			if (m_rt.getvec(VectorTreeDepth, vec) < RTTP_NOER) return RTTP_MNER;
			vec->getnm(OldName); return RTTP_RGHT; }

	erttp regq(vector<vctrdpttp> &VectorTreeDepth, vctrmnttp &SubVectorID,
		string VectorName = "")
		{ cfgvec *vec;
			if (m_rt.getvec(VectorTreeDepth, vec) < RTTP_NOER) return RTTP_MNER;
			return vec->regq(SubVectorID, VectorName); }
	erttp sqz(vector<vctrdpttp> &OutputVectors, vector<vctrmnttp> &OldVectors,
		vector<vctrdpttp> &VectorTreeDepth)
		{ cfgvec *vec;
			if (m_rt.getvec(VectorTreeDepth, vec) < RTTP_NOER) return RTTP_MNER;
			return vec->sqz(OutputVectors, OldVectors); }
	erttp wlk(vector<vctrdpttp> &VectorTreeDepth, vctrdpttp &VectorInfo) const
	{ cfgvec *vec;
		if (m_rt.getvec(VectorTreeDepth, vec) < RTTP_NOER) return RTTP_MNER;
		return vec->wlk(VectorInfo); }
};

/*	g++ seems not to fully support seperation compilation module,
	so all the definitions related to templates are put here	*/

/*	element registration or reinitialization
Return Values:
	RTTP_NOER:	no value set yet, but input values may already been polluted
	RTTP_RGHT:	value got
*/
template <class EleType>
erttp ezconfig::cfgvec::cfgele<EleType>::reg(const vctrlvltp lvlmnt)
{
	if (m_elevl.max_size() >= lvlmnt)
	{
		elevltp prsmpl;	//pair sample
		EleType elesmpl;	//element sample

		prsmpl = make_pair(false, elesmpl);
		m_elevl.resize(lvlmnt, prsmpl);
		return RTTP_RGHT;
	}
	else
		return RTTP_CRER;
}

/*	set element's value
Return Values:
	RTTP_CRER:	requested level unsupported
	RTTP_NOER:	value overwritten
	RTTP_RGHT:	value set for the 1st. time
*/
template <class EleType>
erttp ezconfig::cfgvec::cfgele<EleType>::set(const vctrlvltp lvl,
	const EleType vl)
{
	erttp ret = RTTP_RGHT;

	if (m_elevl.size() < lvl + 1)
		return RTTP_CRER;
	if (true == (m_elevl.at(lvl)).first)	//reset
		ret = RTTP_NOER;
	else
		(m_elevl.at(lvl)).first = true;
	(m_elevl.at(lvl)).second = vl;
	return ret;
}

/*	get element's value at a certain level
Return Values:
	RTTP_MNER:	value not set yet
	RTTP_RGHT:	value got
*/
template <class EleType>
erttp ezconfig::cfgvec::cfgele<EleType>::getbsc(vctrlvltp lvl, EleType &vl)
	const
{
	if (false == (m_elevl.at(lvl)).first)
		return RTTP_MNER;
	vl = (m_elevl.at(lvl)).second;
	return RTTP_RGHT;
}

/*	get element's value at the highest possible level
Return Values:
	RTTP_MNER:	no value set yet, but input values may already been polluted
	RTTP_RGHT:	value got
*/
template <class EleType>
erttp ezconfig::cfgvec::cfgele<EleType>::get(vctrlvltp &lvl, EleType &vl) const
{
	if (lvl > m_elevl.size())
		lvl = m_elevl.size();
	for (; lvl > 0; lvl--)
	{
		if (RTTP_RGHT == getbsc(lvl - 1, vl))
		{
			lvl--;
			return RTTP_RGHT;
		}
	}
	return RTTP_MNER;
}


/*	element registration or reinitialization
Return Values:
	RTTP_CRER:	memory failure
	RTTP_MNER:	no corresponding vector
	RTTP_NOER:	already registered
	RTTP_RGHT:	registered
*/
template <typename EleType>
erttp ezconfig::cfgvec::regele(vctrlvltp lvl)
{
	erttp ret;

	if (NULL != m_vecvl)
		ret = RTTP_NOER;
	else
	{
		m_vecvl = new(cfgele<EleType>);
		if (NULL == m_vecvl)
			ret = RTTP_CRER;
		else
			ret = (reinterpret_cast<cfgele<EleType> *>(m_vecvl))->reg(lvl);
	}
	return ret;
}

/*	element release
Return Values:
	RTTP_CRER:	sub-elements in existence
	RTTP_MNER:	no corresponding vector
	RTTP_NOER:	already destroyed or not registered yet
	RTTP_RGHT:	registration destroyed and removed
*/
template <typename EleType>
erttp ezconfig::cfgvec::unregele(void)
{
	cfgvec vc;
	erttp ret = RTTP_RGHT;

	if (m_q.size() != 0)
		ret = RTTP_CRER;
	else
		if (NULL == m_vecvl)
			ret = RTTP_NOER;
		else
			delete(reinterpret_cast<cfgele<EleType> *>(m_vecvl));
	return ret;
}

/*	set element's value
Return Values:
	RTTP_CRER:	requested level unsupported
	RTTP_MNER:	requested vector not found or not registered yet
	RTTP_NOER:	value overwritten
	RTTP_RGHT:	value set for the 1st. time
*/
template <typename EleType>
erttp ezconfig::cfgvec::setele(vctrlvltp lvl, EleType vl)
{
	cfgvec vc;
	erttp ret;

	if (NULL == m_vecvl)
		ret = RTTP_MNER;
	else
	{
		ret = (reinterpret_cast<cfgele<EleType> *>(m_vecvl))->set(lvl, vl);
	}
	return ret;
}

/*	get element's value at a certain level & its name
Input Parameters:
	dp:	the IDs or names of the vectors on the way to the one asked for.
		the upper its position is, or, the closer it's to the root, the latter
		it is in the list.
		cleared if found. the ones remained in the list help keep track of
		where a vector cannot be found.
	lvl:	get the element at this specific level, if this level's in the
		registered boundary. otherwise, i.e., larger than registered,
		the highest set level returned.
	vl:	set if found. otherwise, not set.
	nm:	the element's name returned.
Return Values:
	RTTP_NOER:	value not set yet
	RTTP_RGHT:	value got
*/
template <typename EleType>
erttp ezconfig::cfgvec::getele(vctrlvltp &lvl, EleType &vl, string &nm) const
{
	cfgvec vc;
	erttp ret;

	if (NULL == m_vecvl)
		ret = RTTP_MNER;
	else
	{
		getnm(nm);
		ret = (reinterpret_cast<cfgele<EleType> *>(m_vecvl))->get(lvl, vl);
	}
	return ret;
}

/*	element release
Return Values:
	RTTP_CRER:	sub-elements in existence
	RTTP_MNER:	no corresponding vector
	RTTP_NOER:	root de-registered
	RTTP_RGHT:	registration destroyed and removed
*/
template <typename EleType>
erttp ezconfig::unregele(vector<vctrdpttp> &dpt, vctrdpttp &nf)
{
	cfgvec *vecchld, *vecprnt = NULL;
	vector<vctrdpttp> chld;
	erttp ret;
	vctrmnttp id;
	string nm;

	if (dpt.size() > 0)
	{	//not root, subtract the bottom element from tree
		chld.assign(1, dpt.at(0));
		dpt.erase(dpt.begin());
		if (m_rt.getvec(dpt, vecprnt) < RTTP_NOER)
			return RTTP_MNER;
		vector<vctrdpttp> chldcln(chld);
		if (vecprnt->getvec(chldcln, vecchld) < RTTP_NOER)
			return RTTP_MNER;
	}
	else	//this is root, this invokation is not necessary in fact so far
		if (m_rt.getvec(dpt, vecchld) < RTTP_NOER)
			return RTTP_MNER;
	vecchld->getid(id);
	vecchld->getnm(nm);
	nf = make_pair(id, nm);
	ret = vecchld->unregele<EleType>();
	if (NULL == vecprnt)
		ret = RTTP_NOER;
	else
		if (ret >= RTTP_NOER)
			ret = vecprnt->unregq(chld);
	return ret;
}


class ezcnfgclnt	//configuration client
{
private:
	//file items handling, return value: whether to continue
	bool (*itmf)(string section, string item, string value);
	//command line items handling, return value: whether to continue
	bool (*itml)(string item, string value);
	//file name interactions, return value: whether to continue
	bool (*hlp)(string &FileName);

public:
	void regitm(bool (*ItemFunctionPointer)( string, string, string))
		{ itmf = ItemFunctionPointer; };
	void regitm(bool (*ItemFunctionPointer)( string, string))
		{ itml = ItemFunctionPointer; };
	void reghlp(bool (*HelpFunctionPointer)(string&))
		{ hlp = HelpFunctionPointer; };
	erttp run(string &FileName, string &ErrorMessages);
	erttp run(int argc, char *argv[], string &ErrorMessages);
};

}	/*	namespace ezproject	*/

#endif	/*	EZCONFIG_H	*/

